//
//  NSString-P3CExtensions.m
//  P3CCocoaExtensions
//
//  Created by Pascal on Mon Mar 04 2002.
//  Copyright (c) 2002 P3 Consulting. All rights reserved.
//
/*
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#import "NSString-P3CExtensions.h"
#import <unistd.h>


static const char trailingBytesForUTF8[256] = {
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, 3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5
};

@implementation NSString(P3TrimString)

-(NSString *) trim
{
	return [self stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]]	;
}

@end

@implementation NSString(P3SupprSpaceString)

-(NSString *) supprspace
{
	int	len = [self cStringLength]	;
	char	*buffer = (char *)malloc( len+1 ), *d ;
	const char *s	;
	NSString	*result = nil	;
		
		if (buffer == nil)
			return nil	;
		
		d = buffer	;
		s = [self cString]	;
		
		while (*s)	{
			if (*s != ' ')
				*d++ = *s	;
			++s	;
		}
		*d = '\0'	;
			
		result = [NSString stringWithCString:buffer]	;
		free(buffer)	;
		
		return result	;
}

@end

@implementation NSString(P3EscapeForHTML)

-(NSString *) escapeStringForHTML
{
	int	i, len = [self length], iStart	;
	NSMutableString	*result = [NSMutableString stringWithCapacity:len]	;
	unichar	*buffer = malloc(len * sizeof(unichar) )	;
				
		[self getCharacters:buffer]	;
		
		iStart = 0	;
		for(i=0;i<len;i++)	{
			unichar	c = buffer[i]	;
			  
			if (c < 0x0080)	{
				switch (c)	{
					case	0x0022	:
						if (i > iStart)	{
							[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
						}
						[result appendString:@"&quot;"]	;
						iStart = i + 1	;
						break	;
						
					case	0x0026	:
						if (i > iStart)	{
							[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
						}
						[result appendString:@"&amp;"]	;
						iStart = i + 1	;
						break	;
						
					case	0x003C	:
						if (i > iStart)	{
							[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
						}
						[result appendString:@"&lt;"]	;
						iStart = i + 1	;
						break	;
						
					case	0x003E	:
						if (i > iStart)	{
							[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
						}
						[result appendString:@"&gt;"]	;
						iStart = i + 1	;
						break	;
						
					default	:
						break;
				}
			}
			else	{
				if (i > iStart)	{
					[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
				}
				iStart = i + 1	;
				[result appendString:[NSString stringWithFormat:@"&#%d;",c]]	;
			}
		}
		if (len > iStart)	{
			[self getCharacters:buffer range:NSMakeRange(iStart,len-iStart)]	;
			[result appendString:[NSString stringWithCharacters:buffer length:len-iStart]]	;
		}
		
		free(buffer)	;
		return result	;
}

-(NSString *) unescapeStringFromHTML
{
	int	i, len = [self length], iStart = 0	;
	NSMutableString	*result = [NSMutableString stringWithCapacity:len]	;
	unichar	*buffer = malloc(len * sizeof(unichar) )	;
		
		[self getCharacters:buffer]	;
		for(i=0;i<len;i++)	{
			unichar	c = buffer[i]	;
			unichar	c1 = buffer[i+1]	;
		
			if (c == 0x0026) 	{	// & #
				unichar		a[2]	;
				
				if (c1 == 0x0023)	{
					int	value = 0	;
					iStart = i + 2	;
					c = buffer[iStart]	;
					while (c != 0x003B)	{
						value = (value * 10) + (c - 0x0030)	;
						c = buffer[++iStart]	;
					}
					++iStart	;
					a[0] = value	;
					[result appendString:[NSString stringWithCharacters:a length:1]]	;
				}
				else	{					
					
					if (i > iStart)
						[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
					
					iStart = i	;
					do
						c = buffer[++i]	;
					while (c != 0x003B)	;
					// iStart ... i range of html escape sequence
	
					NSString	*htmlSeq = [NSString stringWithCharacters:buffer+iStart length:i+1-iStart]	;
					iStart = i+1	;
					
					if ([htmlSeq isEqualToString:@"&amp;"])	{
						a[0] = 0x0026	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&quot;"])	{
						a[0] = 0x0022	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&lt;"])	{
						a[0] = 0x003C	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&gt;"])	{
						a[0] = 0x003E	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&nbsp;"])	{
						a[0] = 0x00A0	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&iexcl;"])	{
						a[0] = 0x00A1	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&cent;"])	{
						a[0] = 0x00A2	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&pound;"])	{
						a[0] = 0x00A3	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&curren;"])	{
						a[0] = 0x00A4	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&yen;"])	{
						a[0] = 0x00A5	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&brvbar;"])	{
						a[0] = 0x00A6	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&sect;"])	{
						a[0] = 0x00A7	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&uml;"])	{
						a[0] = 0x00A8	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&copy;"])	{
						a[0] = 0x00A9	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&ordf;"])	{
						a[0] = 0x00AA	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&laquo;"])	{
						a[0] = 0x00AB	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&not;"])	{
						a[0] = 0x00AC	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&shy;"])	{
						a[0] = 0x00AD	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&reg;"])	{
						a[0] = 0x00AE	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&macr;"])	{
						a[0] = 0x00AF	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&deg;"])	{
						a[0] = 0x00B0	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&plusmn;"])	{
						a[0] = 0x00B1	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&sup2;"])	{
						a[0] = 0x00B2	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&sup3;"])	{
						a[0] = 0x00B3	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&acute;"])	{
						a[0] = 0x00B4	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&micro;"])	{
						a[0] = 0x00B5	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&para;"])	{
						a[0] = 0x00B6	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&cedil;"])	{
						a[0] = 0x00B6	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&middot;"])	{
						a[0] = 0x00B7	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&cedil;"])	{
						a[0] = 0x00B8	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&sup1;"])	{
						a[0] = 0x00B9	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&ordm;"])	{
						a[0] = 0x00BA	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&raquo;"])	{
						a[0] = 0x00BB	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&frac14;"])	{
						a[0] = 0x00BC	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&frac12;"])	{
						a[0] = 0x00BD	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&frac34;"])	{
						a[0] = 0x00BE	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&iquest;"])	{
						a[0] = 0x00BF	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Agrave;"])	{
						a[0] = 0x00C0	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Aacute;"])	{
						a[0] = 0x00C1	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Acirc;"])	{
						a[0] = 0x00C2	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Atilde;"])	{
						a[0] = 0x00C3	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Auml;"])	{
						a[0] = 0x00C4	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Aring;"])	{
						a[0] = 0x00C5	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&AElig;"])	{
						a[0] = 0x00C6	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Ccedil;"])	{
						a[0] = 0x00C7	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Egrave;"])	{
						a[0] = 0x00C8	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Eacute;"])	{
						a[0] = 0x00C9	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Ecirc;"])	{
						a[0] = 0x00CA	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Euml;"])	{
						a[0] = 0x00CB	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Igrave;"])	{
						a[0] = 0x00CC	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Iacute;"])	{
						a[0] = 0x00CD	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Icirc;"])	{
						a[0] = 0x00CE	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Iuml;"])	{
						a[0] = 0x00CF	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&ETH;"])	{
						a[0] = 0x00D0	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Ntilde;"])	{
						a[0] = 0x00D1	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Ograve;"])	{
						a[0] = 0x00D2	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Oacute;"])	{
						a[0] = 0x00D3	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Ocirc;"])	{
						a[0] = 0x00D4	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Otilde;"])	{
						a[0] = 0x00D5	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Ouml;"])	{
						a[0] = 0x00D6	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Otimes;"])	{
						a[0] = 0x00D7	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Oslash;"])	{
						a[0] = 0x00D8	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Ugrave;"])	{
						a[0] = 0x00D9	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Uacute;"])	{
						a[0] = 0x00DA	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Ucirc;"])	{
						a[0] = 0x00DB	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Uuml;"])	{
						a[0] = 0x00DC	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&Yacute;"])	{
						a[0] = 0x00DD	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&THORN;"])	{
						a[0] = 0x00DE	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&szlig;"])	{
						a[0] = 0x00DF	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&agrave;"])	{
						a[0] = 0x00E0	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&aacute;"])	{
						a[0] = 0x00E1	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&acirc;"])	{
						a[0] = 0x00E2	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&atilde;"])	{
						a[0] = 0x00E3	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&auml;"])	{
						a[0] = 0x00E4	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&aring;"])	{
						a[0] = 0x00E5	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&aelig;"])	{
						a[0] = 0x00E6	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&ccedil;"])	{
						a[0] = 0x00E7	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&egrave;"])	{
						a[0] = 0x00E8	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&eacute;"])	{
						a[0] = 0x00E9	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&ecirc;"])	{
						a[0] = 0x00EA	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&euml;"])	{
						a[0] = 0x00EB	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&igrave;"])	{
						a[0] = 0x00EC	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&iacute;"])	{
						a[0] = 0x00ED	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&icirc;"])	{
						a[0] = 0x00EE	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&iuml;"])	{
						a[0] = 0x00EF	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&eth;"])	{
						a[0] = 0x00F0	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&ntilde;"])	{
						a[0] = 0x00F1	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&ograve;"])	{
						a[0] = 0x00F2	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&oacute;"])	{
						a[0] = 0x00F3	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&ocirc;"])	{
						a[0] = 0x00F4	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&otilde;"])	{
						a[0] = 0x00F5	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&ouml;"])	{
						a[0] = 0x00F6	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&divide;"])	{
						a[0] = 0x00F7	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&oslash;"])	{
						a[0] = 0x00F8	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&ugrave;"])	{
						a[0] = 0x00F9	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&uacute;"])	{
						a[0] = 0x00FA	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&ucirc;"])	{
						a[0] = 0x00FB	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&uuml;"])	{
						a[0] = 0x00FC	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&yacute;"])	{
						a[0] = 0x00FD	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&thorn;"])	{
						a[0] = 0x00FE	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	if ([htmlSeq isEqualToString:@"&yuml;"])	{
						a[0] = 0x00FF	;
						[result appendString:[NSString stringWithCharacters:a length:1]]	;
					}
					else	{
						[result appendString:htmlSeq]	;
					}
				}
			}
		}
		if (len > iStart)	{
			[result appendString:[NSString stringWithCharacters:buffer+iStart length:len-iStart]]	;
		}
		
		free(buffer)	;
		return result	;
}

@end

@implementation NSString(P3EscapeForSQL)

-(NSString *) escapeStringForSQL
{
	int	i, len = [self length], iStart = 0	;
	NSMutableString	*result = [NSMutableString stringWithCapacity:len]	;
	unichar	*buffer = malloc(len * sizeof(unichar) )	;

		[self getCharacters:buffer]	;
		for(i=0;i<len;i++)	{
			unichar	c = buffer[i]	;

			switch(c)	{
				case 0x0027	:	// '
					if (i > iStart)	{
						[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
					}
					[result appendString:@"\\'"]	;
					iStart = i + 1	;
					break	;
					
				case 0x0022	:	// "
					if (i > iStart)	{
						[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
					}
					[result appendString:@"\\\""]	;
					iStart = i + 1	;
					break	;
					
				case 0x0009	:
					if (i > iStart)	{
						[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
					}
					[result appendString:@"\\t"]	;
					iStart = i + 1	;
					break	;
					
				case 0x000A	:
					if (i > iStart)	{
						[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
					}
					[result appendString:@"\\n"]	;
					iStart = i + 1	;
					break	;
					
				case 0x000B	:
					if (i > iStart)	{
						[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
					}
					[result appendString:@"\\f"]	;
					iStart = i + 1	;
					break	;
					
				case 0x000D	:
					if (i > iStart)	{
						[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
					}
					[result appendString:@"\\r"]	;
					iStart = i + 1	;
					break	;					
					
				case 0x005C	:	// backslash
					if (i > iStart)	{
						[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
					}
					[result appendString:@"\\\\"]	;
					iStart = i + 1	;
					break	;					
			}
		}
		if (len > iStart)	{
			[result appendString:[NSString stringWithCharacters:buffer+iStart length:len-iStart]]	;
		}
		
		free(buffer)	;
		return result	;
}

- (NSString *)unescapeStringFromSQL
{
	int	i, len = [self length], iStart = 0	;
	NSMutableString	*result = [NSMutableString stringWithCapacity:len]	;
	unichar	*buffer = malloc(len * sizeof(unichar) )	;

		[self getCharacters:buffer]	;
		for(i=0;i<len;i++)	{
			unichar	c = buffer[i]	;
			
			if (c == 0x005C)	{
				++i	;
				c = buffer[i]	;
				
				switch(c)	{
					case	0x0027	:	// '
						buffer[i-1] = 0x0027	;
						[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
						iStart = i + 1	;						
						break	;
					case	0x0022	:	// "	
						buffer[i-1] = 0x0022	;
						[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
						iStart = i + 1	;						
						break	;
					case	0x0066	:	// f	
						buffer[i-1] = 0x000B	;
						[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
						iStart = i + 1	;						
						break	;
					case	0x006E	:	// n	
						buffer[i-1] = 0x000A	;
						[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
						iStart = i + 1	;						
						break	;
					case	0x0072	:	// r	
						buffer[i-1] = 0x000C	;
						[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
						iStart = i + 1	;						
						break	;
					case	0x0074	:	// t	
						buffer[i] = 0x0009	;
						[result appendString:[NSString stringWithCharacters:buffer+iStart length:i-iStart]]	;
						iStart = i + 1	;						
						break	;
				}
			}
		}
		if (len > iStart)	{
			[result appendString:[NSString stringWithCharacters:buffer+iStart length:len-iStart]]	;
		}
		
		free(buffer)	;
		
		return result	;
}

@end

@implementation NSString(UTF8length)

-(int) UTF8length
{
	const char	*utf8 = [self UTF8String]	;
	int			len = 0	;
	
		while (*utf8)	{
			int	c = (int)*utf8	;
			
			++len	;
			utf8 += trailingBytesForUTF8[c]	;
			++utf8	;
		}
		return len	;
}

@end


@implementation NSString(P3EscapeForC)

//	\n -> \\n, \t -> \\t
-(NSString *) escapeStringForC
{
	const char	*utf8 = [self UTF8String]	;
	char		*outBuffer = malloc( 2 * strlen(utf8) + 1)	;
	const char	*source = utf8		;
	char		*dest = outBuffer	;
	NSString	*result	;
		
		while (*source)	{
			if(*source == '\n')	{
				*dest++ = '\\'	;
				*dest++ = 'n'	;
			}
			else	if(*source == '\t')	{
				*dest++ = '\\'	;
				*dest++ = 't'	;
			}
			else	if(*source == '\r')	{
				*dest++ = '\\'	;
				*dest++ = 'r'	;
			}
			else
				*dest++ = *source++	;
		}
		
		*dest = '\0'	;
		result = [NSString stringWithUTF8String:outBuffer]	;

		free(outBuffer)	;
		
		return result	;
}

-(NSString *) unescapeStringFromC
{
	const char	*utf8 = [self UTF8String]	;
	char		*outBuffer = malloc( strlen(utf8) + 1)	;
	const char	*source = utf8		;
	char		*dest = outBuffer	;
	NSString	*result	;
	
		while (*source)	{
			if (*source == '\\')	{
				++source	;
				if (*source == 'n')
					*dest++ = '\n'	;
				else	if (*source == 'r')
					*dest++ = '\r'	;
				else	if (*source == 't')
					*dest++ = '\t'	;
				++source	;
			}
			else
				*dest++ = *source++	;
		}
		
		*dest = '\0'	;
		result = [NSString stringWithUTF8String:outBuffer]	;

		free(outBuffer)	;
		
		return result	;
}

@end

@implementation NSString(P3EscapeForXML)

-(NSString *) escapeStringForXML
{
	int				len = [self length]	;
	NSMutableString	*result = [NSMutableString stringWithCapacity:len]	;
	const char	*utf8 = [self UTF8String]	;
	char		*buffer = malloc( strlen(utf8) + 1)	;
		strcpy(buffer,utf8)	;

	char				*start = buffer		;
	char				*current = buffer	;
	unsigned char		c = (unsigned char)*current		;
	
		while (c != '\0')	{
			switch (c)	{
				case	0x0D	:
					c = 0x0A	;
					break;
					
				case	0x22	:
					if (current > start)	{
						*current = '\0'	;
						[result appendString:[NSString stringWithUTF8String:start]]	;
					}
					[result appendString:@"&quot;"]	;
					start = current + 1	;
					break	;
					
				case	0x27	:
					if (current > start)	{
						*current = '\0'	;
						[result appendString:[NSString stringWithUTF8String:start]]	;
					}
					[result appendString:@"&apos;"]	;
					start = current + 1	;
					break	;
					
				case	0x26	:
					if (current > start)	{
						*current = '\0'	;
						[result appendString:[NSString stringWithUTF8String:start]]	;
					}
					[result appendString:@"&amp;"]	;
					start = current + 1	;
					break	;
					
				case	0x3C	:
					if (current > start)	{
						*current = '\0'	;
						[result appendString:[NSString stringWithUTF8String:start]]	;
					}
					[result appendString:@"&lt;"]	;
					start = current + 1	;
					break	;
					
				case	0x3E	:
					if (current > start)	{
						*current = '\0'	;
						[result appendString:[NSString stringWithUTF8String:start]]	;
					}
					[result appendString:@"&gt;"]	;
					start = current + 1	;
					break	;
			}

			current = current + 1 + trailingBytesForUTF8[c]	;
			c = (unsigned char)*current	;
		}
		if (*start != '\0')	{
			[result appendString:[NSString stringWithUTF8String:start]]	;
		}
		
		free(buffer)	;
		
		return result	;
}

@end

@implementation NSString(ABUtils)
-(NSString *) extractABLabel
{
	NSString	*result	;
	UniChar	*buffer	;
	
		buffer = malloc([self length] * sizeof(UniChar))	;
		if ([self hasPrefix:@"_$!<"])	{
			[self getCharacters:buffer range:NSMakeRange(4,[self length]-4)]	;
		
			result = [NSString stringWithCharacters:buffer length:[self length]-4]	;
			
			if ([result hasSuffix:@">!$_"])	{
				[result getCharacters:buffer range:NSMakeRange(0,[result length]-4)]	;
				result = [NSString stringWithCharacters:buffer length:[result length]-4]	;
			}
		}
		else
			result = [NSString stringWithString:self]	;
		free(buffer)	;	
			
		return result	;
}

@end

@implementation NSString(RFC2396Support)
-(NSString *) stringByAddingPercentEscapes
{
		return [ (NSString *)CFURLCreateStringByAddingPercentEscapes(NULL, (CFStringRef)self, NULL, NULL, 
					CFStringConvertNSStringEncodingToEncoding(NSASCIIStringEncoding)) autorelease];
}

-(NSString *) stringByReplacingPercentEscapes
{
		return [ (NSString *)CFURLCreateStringByReplacingPercentEscapes(kCFAllocatorDefault, (CFStringRef)self, CFSTR("")) autorelease];
}

@end

@implementation NSString(ReadLine)

- (NSArray *) readFieldsForSeparator:(NSString *)fieldSep lineEndIndex:(unsigned *)lineEndIndex
{
	unsigned	startIndex, contentsEndIndex	;
	NSArray		*result = nil	;
	NSString	*str	;
	
		NS_DURING
			[self getLineStart:&startIndex end:lineEndIndex contentsEnd:&contentsEndIndex forRange:NSMakeRange(0,0)]	;

			unichar	*buffer = (unichar *)malloc( (contentsEndIndex - startIndex) * sizeof( unichar ) )	;
			[self getCharacters:buffer range:NSMakeRange(startIndex,contentsEndIndex)]	;
			str = [NSString stringWithCharacters:buffer length:contentsEndIndex-startIndex]	;
			result = [str componentsSeparatedByString:fieldSep]	;
			free(buffer)	;
		NS_HANDLER
		NS_ENDHANDLER
		
		return result	;
}

@end


@implementation NSString(TabExport)

- (BOOL)containsString:(NSString *)searchString options:(unsigned int)mask;
{
    return !searchString || [searchString length] == 0 || [self rangeOfString:searchString options:mask].length > 0;
}

- (BOOL)containsString:(NSString *)searchString;
{
    return !searchString || [searchString length] == 0 || [self rangeOfString:searchString].length > 0;
}

- (NSString *)stringByReplacingReturns:(NSString *)byString
{
    return [[self stringByReplacingString:@"\n" by:byString] stringByReplacingString:@"\r" by:byString];
}

- (NSString *)stringByReplacingString:(NSString *)removeString by:(NSString *)byString
{
    NSArray *lines;
    NSMutableString *newString;
    NSString *returnValue;
    unsigned int lineIndex, lineCount;

		if (![self containsString:removeString])
			return [[self copy] autorelease];
		newString = [[NSMutableString alloc] initWithCapacity:[self length]];
		lines = [self componentsSeparatedByString:removeString];
		lineCount = [lines count];
		for (lineIndex = 0; lineIndex < lineCount-1; lineIndex++)
			[newString appendFormat:@"%@%@",[lines objectAtIndex:lineIndex],byString];
		[newString appendString:[lines objectAtIndex:lineCount-1]];
		returnValue = [newString copy];
		[newString release];

		return [returnValue autorelease];
}

@end

@implementation NSString(P3ParseACL)

//	parse a postgreSQL ACL list {=,user=access_rights,...}
//	returns a dict with users's name as keys
//	empty user ("everybody else" user generates 2 keys "*" and "PUBLIC")
 
-(NSDictionary *)parseACL;
{
	char	*buffer = malloc( sizeof(char) * ([self cStringLength]+1) ), *s = buffer, *sKey, *sValue	;
	NSMutableDictionary *dict = [NSMutableDictionary dictionary]	;
	
		strcpy(buffer,[self cString])	;
		
		if (*s && (*s == '{'))	{
			sKey = ++s	;
			while (*s && (*s != '}'))	{

				while (*s && (*s != '='))	++s	;

				if (*s == '=')		{
					*s = '\0'	;
					sValue = ++s	;

					while (*s && (*s != ',') && (*s != '}'))	++s	;

					if (*s)	{
						*s = '\0'	;
						
						if (strlen(sKey) == 0)	{
							[dict setObject:[NSString stringWithCString:sValue] forKey:@"*"]	;
						}
						else
							[dict setObject:[NSString stringWithCString:sValue] forKey:[NSString stringWithCString:sKey]]	;
						
						sKey = ++s	;
					}
				}
				else 
					break;
			}
		}
		
		free(buffer);
		return dict	;
}

+(NSString *)fromACLDict:(NSDictionary *)inDict
{
	NSArray	*keys = [inDict allKeys]	;
	int	i, n = [keys count]	;
		
		if (n == 0)
			return [NSString stringWithString:@""]	;
			
	NSMutableString	*string = [NSMutableString stringWithString:@"{"]	;
	
		for(i=0;i<n-1;i++)	{
			if ([[keys objectAtIndex:i] isEqualToString:@"*"])	{
				[string appendString:@"="]	;
				[string appendString:[inDict objectForKey:[keys objectAtIndex:i]]]	;
			}
			else	{
				[string appendString:[keys objectAtIndex:i]]	;
				[string appendString:@"="]	;
				[string appendString:[inDict objectForKey:[keys objectAtIndex:i]]]	;
			}
			[string appendString:@","]	;
		}
		if (n-1 >= 0)	{
			if ([[keys objectAtIndex:n-1] isEqualToString:@"*"])	{
				[string appendString:@"="]	;
				[string appendString:[inDict objectForKey:[keys objectAtIndex:n-1]]]	;
			}
			else	{
				[string appendString:[keys objectAtIndex:i]]	;
				[string appendString:@"="]	;
				[string appendString:[inDict objectForKey:[keys objectAtIndex:n-1]]]	;
			}
		}
		 
		[string appendString:@"}"]	;
		
		return [NSString stringWithString:string]	;
}

@end

@implementation NSString(P3ParseIntArray)

-(NSArray *)parseIntArray
{
	char	*buffer = malloc( sizeof(char) * ([self cStringLength]+1) ), *s = buffer, *sValue	;
	NSMutableArray *array = [NSMutableArray array]	;
	int		intValue	;
	
		strcpy(buffer,[self cString])	;
		
		if (*s++ == '{')	{
			while (*s)	{
				sValue = s	;
				while (*s && (*s != ',') && (*s != '}'))	++s	;
				
				*s = '\0'	;
				sscanf(sValue,"%d",&intValue)	;
				[array addObject:[NSNumber numberWithInt:intValue]]	;
				++s	;
			}
		}
		
		if (*sValue)	{
		
		}
		
		free(buffer);
		return array	;
}

+(NSString *)fromIntArray:(NSArray *)inArray startList:(NSString *)inStart endList:(NSString *)inEnd
{
	int	i, n = [inArray count]	;
		
		if (n == 0)
			return [NSString stringWithString:@""]	;
			
	NSMutableString	*string = [NSMutableString stringWithString:inStart]	;
	
		for(i=0;i<n-1;i++)	{
			[string appendString:[NSString stringWithFormat:@"%d",[[inArray objectAtIndex:i] intValue]]]	;
			[string appendString:@","]	;
		}
		if (n-1 >= 0)	{
			[string appendString:[NSString stringWithFormat:@"%d",[[inArray objectAtIndex:n-1] intValue]]]	;
		}
		 
		[string appendString:inEnd]	;
		
		return [NSString stringWithString:string]	;
}

+(NSString *)fromIntArray:(NSArray *)inArray
{
		return [NSString fromIntArray:inArray startList:@"{" endList:@"}"]	;
}

@end

@implementation NSString(P3ParseLongLongArray)

-(NSArray *)parseLongLongArray
{
	char	*buffer = malloc( sizeof(char) * ([self cStringLength]+1) ), *s = buffer, *sValue	;
	NSMutableArray *array = [NSMutableArray array]	;
	long long 		longLongValue	;
	
		strcpy(buffer,[self cString])	;
		
		if (*s++ == '{')	{
			while (*s)	{
				sValue = s	;
				while (*s && (*s != ',') && (*s != '}'))	++s	;
				
				*s = '\0'	;
				sscanf(sValue,"%lld",&longLongValue)	;
				[array addObject:[NSNumber numberWithLongLong:longLongValue]]	;
				++s	;
			}
		}
		
		free(buffer);
		return array	;
}

+(NSString *)fromLongLongArray:(NSArray *)inArray startList:(NSString *)inStart endList:(NSString *)inEnd
{
	int	i, n = [inArray count]	;
		
		if (n == 0)
			return [NSString stringWithString:@""]	;
			
	NSMutableString	*string = [NSMutableString stringWithString:inStart]	;
	
		for(i=0;i<n-1;i++)	{
			[string appendString:[NSString stringWithFormat:@"%lld",[[inArray objectAtIndex:i] longLongValue]]]	;
			[string appendString:@","]	;
		}
		if (n-1 >= 0)	{
			[string appendString:[NSString stringWithFormat:@"%d",[[inArray objectAtIndex:n-1] longLongValue]]]	;
		}
		 
		[string appendString:inEnd]	;
		
		return [NSString stringWithString:string]	;
}

+(NSString *)fromLongLongArray:(NSArray *)inArray
{
		return [NSString fromLongLongArray:inArray startList:@"{" endList:@"}"]	;
}

@end

@implementation NSString(Dictionary)

-(NSDictionary *) asDictionary
{
		[self propertyList]	;
/*
	char	tempFileName[256]	;		// mktemp requires a writable string buffer
		strcpy(tempFileName,"/tmp/P3CCocoaExtensions.XXXXXX")	;
	NSString	*s = [NSString stringWithCString:mktemp(tempFileName)]	;
	
		if (![self writeToFile:s atomically:YES])
			return nil	;

		// tmp files are deleted at reboot time no need to delete it from here
			
		return [NSDictionary dictionaryWithContentsOfFile:s]	;
*/
}

@end

@implementation NSString (NSStringTextFinding)

- (NSRange)findString:(NSString *)string selectedRange:(NSRange)selectedRange options:(unsigned)options wrap:(BOOL)wrap {
    BOOL forwards = (options & NSBackwardsSearch) == 0;
    unsigned length = [self length];
    NSRange searchRange, range;

    if (forwards) {
		searchRange.location = NSMaxRange(selectedRange);
		searchRange.length = length - searchRange.location;
		range = [self rangeOfString:string options:options range:searchRange];
		if ((range.length == 0) && wrap) {	/* If not found look at the first part of the string */
			searchRange.location = 0;
			searchRange.length = selectedRange.location;
			range = [self rangeOfString:string options:options range:searchRange];
		}
    } else {
		searchRange.location = 0;
		searchRange.length = selectedRange.location;
		range = [self rangeOfString:string options:options range:searchRange];
		if ((range.length == 0) && wrap) {
			searchRange.location = NSMaxRange(selectedRange);
			searchRange.length = length - searchRange.location;
			range = [self rangeOfString:string options:options range:searchRange];
		}
    }
    return range;
}        

@end


@implementation NSString(BoolValue)

- (BOOL)boolValue
{
		NS_DURING
			int	i = [self intValue]	;
			
				return (i != 0)	;
		NS_HANDLER
			return [self isEqualToString:@"YES"] ||
					[self isEqualToString:@"yes"] ||
					[self isEqualToString:@"Yes"] ||
					[self isEqualToString:@"Y"] ||
					[self isEqualToString:@"y"] ||
					[self isEqualToString:@"T"] ||
					[self isEqualToString:@"t"] ||
					[self isEqualToString:@"true"] ||
					[self isEqualToString:@"True"] ||
					[self isEqualToString:@"TRUE"] ;
		NS_ENDHANDLER
}

@end


@implementation NSString(Quote)

- (NSString *)quote
{
		return [NSString stringWithFormat:@"\"%@\"",self]	;
}

@end
